/*
 * Copyright (c) 2008 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef _VELODYNELIB_H
#define _VELODYNELIB_H

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Constants
 */

/* Properties common to both models */
#define VELODYNE_NUM_LASERS_PER_FIRING 32
#define VELODYNE_NUM_FIRINGS_PER_PACKET 12
/* number of lasers per packet: 12 x 32 = 384 */
#define VELODYNE_NUM_LASERS_PER_PACKET (VELODYNE_NUM_LASERS_PER_FIRING \
                                        * VELODYNE_NUM_FIRINGS_PER_PACKET)
#define VELODYNE_PAYLOAD_SIZE 1206

/* Properties specific to each model */
#define VELODYNE_NUM_LASERS_32 32
#define VELODYNE_NUM_LASERS_64 64
/* number of shots of a laser beam collected in a packet */
#define VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_32 12
#define VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_64 6
#define VELODYNE_MEAN_NUM_PACKETS_PER_REVOLUTION_32 180
#define VELODYNE_MEAN_NUM_PACKETS_PER_REVOLUTION_64 730
#define VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_32 200
#define VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_64 750

/* Properties only applicable to the HDL-64 */
#define VELODYNE_MAX_COLUMN_OFFSET_64 250

/**
 * Enumerations
 */

/* Blocks and correspondind IDs, fixed to 0xEEFF in the HDL-32 */
typedef enum VELODYNE_BLOCK {
	VELODYNE_UPPER = 0xEEFF,
	VELODYNE_LOWER = 0xDDFF
} VELODYNE_BLOCK;

/* Interface type */
typedef enum VELODYNE_TYPE {
	VELODYNE_FILE,
	VELODYNE_NETWORK
} VELODYNE_TYPE;

/* Model: HDL-32 or HDL-64 */
typedef enum VELODYNE_MODEL {
	VELODYNE_32,
	VELODYNE_64
} VELODYNE_MODEL;

/**
 * Interface structure
 */
typedef struct VELODYNE_INTERFACE {
	int fd; /* socket file descriptor for the network case */
	void *pcap; /* libpcap descriptor for the file case */
	VELODYNE_TYPE type;
	VELODYNE_MODEL model;
} VELODYNE_INTERFACE;

/**
 * Raw packet and other low-level data structures
 */

/* Laser, 3 bytes */
typedef struct VELODYNE_LASER {
	u_int16_t distance; /* 2mm increments */
	u_int8_t intensity; /* 0-255 */
} __attribute__((packed)) VELODYNE_LASER;

/* Firing, 100 bytes */
typedef struct VELODYNE_FIRING {
	u_int16_t block_id; /* it's not possible to use VELODYNE_BLOCK here due to the
	                       packed attribute */
	u_int16_t rotational_pos;	/* 0-35999 */
	VELODYNE_LASER lasers[VELODYNE_NUM_LASERS_PER_FIRING];
} __attribute__((packed)) VELODYNE_FIRING;

/* Raw packet, VELODYNE_PAYLOAD_SIZE bytes */
typedef struct VELODYNE_RAW_PACKET {
	VELODYNE_FIRING firings[VELODYNE_NUM_FIRINGS_PER_PACKET];
	u_int8_t status[6]; /* status data whose meaning is specific to each model */
} __attribute__((packed)) VELODYNE_RAW_PACKET;

/**
 * High-level data structures
 */

typedef struct VELODYNE_PACKET {
	struct timeval ts; /* timestamp */
	VELODYNE_RAW_PACKET data;
} VELODYNE_PACKET;

typedef enum VELODYNE_POINT_STATUS {
	VELODYNE_BAD_POINT,
	VELODYNE_GOOD_POINT
} VELODYNE_POINT_STATUS;

typedef struct VELODYNE_CARTESIAN_POINT {
	VELODYNE_POINT_STATUS status;
	float x, y, z;
	u_int8_t intensity;
} VELODYNE_CARTESIAN_POINT;

/* Packet of Cartesian points */
typedef struct VELODYNE_CARTESIAN_PACKET {
	struct timeval ts;
	VELODYNE_CARTESIAN_POINT points[VELODYNE_NUM_LASERS_PER_PACKET];
} VELODYNE_CARTESIAN_PACKET;

/**
 * Errors
 */
#define velodyneErrNum(id, n) ((((id) & 0xFFFF) << 16) | ((n) & 0xFFFF))

#define velodyneId 5420

#define VELODYNE_OK 0
#define VELODYNE_ERR_MODEL velodyneErrNum(velodyneId, 1)
#define VELODYNE_ERR_FOV   velodyneErrNum(velodyneId, 2)

#ifdef HAVE_POCOLIBS
#define M_velodyneLib 5420
#define S_velodyneLib_NO_MEMORY       ((M_velodyneLib << 16) | 1)
#define S_velodyneLib_INVALID_DEVICE  ((M_velodyneLib << 16) | 2)
#define S_velodyneLib_FILE_OPEN_ERROR ((M_velodyneLib << 16) | 3)
#define S_velodyneLib_READ_ERROR      ((M_velodyneLib << 16) | 4)
#endif

/**
 * Prototypes
 */
extern VELODYNE_INTERFACE *velodyne_open_file(VELODYNE_MODEL, const char *, int,
                                              double, double);
extern VELODYNE_INTERFACE *velodyne_open_network(VELODYNE_MODEL, int, double,
                                                 double);
extern void velodyne_init_calibration_values(VELODYNE_MODEL);
extern void velodyne_init_fov_filter(VELODYNE_MODEL, int, double, double);
extern int velodyne_add_laser_to_fov_filter(VELODYNE_MODEL, int);
extern void velodyne_convert_heading_sector_to_original_reference_frame(
		VELODYNE_MODEL, double, double, int *, int *, int *);
extern void velodyne_get_organization_values(VELODYNE_MODEL, int *, int *);
extern void velodyne_close(VELODYNE_INTERFACE *);
extern int velodyne_read_packet(const VELODYNE_INTERFACE *, VELODYNE_PACKET *);
extern void velodyne_to_cartesian_packet(
		VELODYNE_MODEL, const VELODYNE_PACKET *, VELODYNE_CARTESIAN_PACKET *);

/**
 * Debug functions
 */
extern void velodyne_print_packet(const VELODYNE_PACKET *);
extern void velodyne_print_cartesian_packet(const VELODYNE_CARTESIAN_PACKET *);

#ifdef __cplusplus
}
#endif

#endif
